#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var lodash = require('lodash');
var path = require('path');
var minimist = _interopDefault(require('minimist'));

var help = `
CouchDB JWT Auth Server

  A small, standalone server for generating JSON Web Tokens for use with
  CouchDB. This service requires couch_jwt_auth[1] to be installed on the
  CouchDB server.

  [0]: https://github.com/BeneathTheInk/couchdb-jwt-auth-server
  [1]: https://github.com/softapalvelin/couch_jwt_auth

Usage:

  $ couchdb-jwt [OPTIONS]

  -c, --config <file>     A config file with your options. It's a JSON object.
  --couchdb <url>         URL to the CouchDB server that manages users and has
                          couch_jwt_auth installed.
  --secret <secret>       The secret used to sign JWTs. This should match the
                          raw value CouchDB server has set for jwt_auth.hs_secret
                          (Note: hs_secret is base64 encoded in CouchDB's config)
  --expiresIn <exp>       Time in seconds for JWT expiration. Default is 300 (5 min)
  --session.store <name>  The library to use for storing session data. There are
                          two built in options: memory and couch. Additional
                          session options can be passed using the dot syntax.
  --endpoint <ep>         The web server mount path. Defaults to '/'.
  --port <port>           The port to start the HTTP server on. Defaults to 3000.
  -h, --help              Show this message.
  -v, --version           Print the currently installed version of this server.
`;

// using standard require so rollup doesn't include it
const createApp = require("./");

let argv = minimist(process.argv.slice(2), {
  string: [],
  boolean: ['help', 'version', 'production'],
  alias: {
    h: 'help', H: 'help',
    v: 'version', V: 'version',
    c: "config"
  }
});

if (argv.help) {
  console.log('  ' + help.replace(/\n/g, '\n  '));
  process.exit(0);
}

if (argv.version) {
  let pkg = require('./package.json');
  console.log('%s %s', pkg.name, pkg.version || 'edge');
  process.exit(0);
}

if (argv.config) {
  let cfile = typeof argv.config === "string" && argv.config ? argv.config : "config.json";
  lodash.defaultsDeep(argv, require(path.resolve(cfile)));
}

process.env.NODE_ENV = argv.production ? 'production' : process.env.NODE_ENV || 'development';

function panic(e) {
  console.error(e.stack || e);
  process.exit(1);
}

const VALID = ['algorithms', 'couchdb', 'endpoint', 'expiresIn', 'secret', 'session'];

const app = createApp(lodash.pick(argv, VALID));

app.setup().then(() => {
  const server = app.listen(argv.port || 3000, '127.0.0.1', () => {
    const addr = server.address();
    console.log('HTTP server listening at http://%s:%s', addr.address, addr.port);
  });

  server.on('error', panic);
}).catch(panic);
