'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var lodash = require('lodash');
var PouchDB = _interopDefault(require('pouchdb'));
var url = require('url');

function getCouchDbOptions(opts) {
  var href = opts && (opts.href || opts.baseUrl);

  if (opts == null) opts = {};else if (typeof opts === "string") {
    href = opts;
    opts = {};
  } else if (typeof opts !== "object") {
    throw new Error("Expecting an object or string for connection options.");
  } else {
    opts = lodash.clone(opts);
  }

  // parse the href as a string
  if (href) {
    href = url.parse(href, true, true);

    // merge in stuff from the query
    lodash.defaults(opts, href.query);
  }

  // drop useless properties
  href = lodash.omit(href, "host", "href", "url", "search", "query", "hash");

  // no hostname, we assume default couchdb
  if (href.hostname == null) {
    lodash.defaults(href, {
      hostname: "localhost",
      port: 5984
    });
  }

  // other defaults
  href = lodash.defaults(href, {
    protocol: "http:"
  });

  // parse authentication information
  if (href.auth) opts.auth = href.auth;
  if (typeof opts.auth === "string") {
    let _auth = opts.auth.split(":");
    opts.auth = { username: _auth[0] || "", password: _auth[1] || "" };
  }

  // generate an auth-less href
  opts.baseUrl = url.format(lodash.omit(href, "auth"));
  if (opts.baseUrl.substr(-1) === "/") opts.baseUrl = opts.baseUrl.substr(0, opts.baseUrl.length - 1);

  return opts;
}

var asyncToGenerator = function (fn) {
  return function () {
    var gen = fn.apply(this, arguments);
    return new Promise(function (resolve, reject) {
      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          return Promise.resolve(value).then(function (value) {
            step("next", value);
          }, function (err) {
            step("throw", err);
          });
        }
      }

      return step("next");
    });
  };
};











var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

class CouchStore {
  constructor(client) {
    this.client = client;
  }

  add(sid) {
    var _this = this;

    return asyncToGenerator(function* () {
      const exists = yield _this.exists(sid);

      if (!exists) {
        return yield _this.client.post({
          _id: sid,
          created: new Date()
        });
      }
    })();
  }

  _get(sid) {
    var _this2 = this;

    return asyncToGenerator(function* () {
      const res = yield _this2.client.allDocs({
        key: sid,
        limit: 1
      });

      return res.rows.length && res.rows[0].value;
    })();
  }

  exists(sid) {
    var _this3 = this;

    return asyncToGenerator(function* () {
      return !!(yield _this3._get(sid));
    })();
  }

  remove(sid) {
    var _this4 = this;

    return asyncToGenerator(function* () {
      const session = yield _this4._get(sid);

      return yield _this4.client.remove(sid, session.rev);
    })();
  }
}

function createCouchStore(sessOpts, couchOpts) {
  const options = _extends({
    db: "jwt_sessions"
  }, couchOpts, getCouchDbOptions(sessOpts));

  return new CouchStore(new PouchDB(`${options.baseUrl}/${options.db}`, lodash.omit(options, "baseUrl", "db")));
}

module.exports = createCouchStore;
