'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var bodyParser = _interopDefault(require('body-parser'));
var express = _interopDefault(require('express'));
var lodash = require('lodash');
var url = require('url');
var http = require('http');
var superagent = _interopDefault(require('superagent'));
var jwt = _interopDefault(require('jsonwebtoken'));
var crypto = _interopDefault(require('crypto'));

function getCouchDbOptions(opts) {
  var href = opts && (opts.href || opts.baseUrl);

  if (opts == null) opts = {};else if (typeof opts === "string") {
    href = opts;
    opts = {};
  } else if (typeof opts !== "object") {
    throw new Error("Expecting an object or string for connection options.");
  } else {
    opts = lodash.clone(opts);
  }

  // parse the href as a string
  if (href) {
    href = url.parse(href, true, true);

    // merge in stuff from the query
    lodash.defaults(opts, href.query);
  }

  // drop useless properties
  href = lodash.omit(href, "host", "href", "url", "search", "query", "hash");

  // no hostname, we assume default couchdb
  if (href.hostname == null) {
    lodash.defaults(href, {
      hostname: "localhost",
      port: 5984
    });
  }

  // other defaults
  href = lodash.defaults(href, {
    protocol: "http:"
  });

  // parse authentication information
  if (href.auth) opts.auth = href.auth;
  if (typeof opts.auth === "string") {
    let _auth = opts.auth.split(":");
    opts.auth = { username: _auth[0] || "", password: _auth[1] || "" };
  }

  // generate an auth-less href
  opts.baseUrl = url.format(lodash.omit(href, "auth"));
  if (opts.baseUrl.substr(-1) === "/") opts.baseUrl = opts.baseUrl.substr(0, opts.baseUrl.length - 1);

  return opts;
}

var asyncToGenerator = function (fn) {
  return function () {
    var gen = fn.apply(this, arguments);
    return new Promise(function (resolve, reject) {
      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          return Promise.resolve(value).then(function (value) {
            step("next", value);
          }, function (err) {
            step("throw", err);
          });
        }
      }

      return step("next");
    });
  };
};











var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

// fetch/verify existing JWT
let info = (() => {
  var _ref = asyncToGenerator(function* (req, res, next) {
    try {
      const data = yield req.app.validateToken(req.jwt);
      res.ok(data);
    } catch (err) {
      next(err);
    }
  });

  return function info(_x, _x2, _x3) {
    return _ref.apply(this, arguments);
  };
})();

let login = (() => {
  var _ref2 = asyncToGenerator(function* (req, res, next) {
    try {
      const password = req.body.password || req.body.pass;
      const username = req.body.username || req.body.name;
      const result = yield req.app.login(username, password, req.body);
      res.ok(result);
    } catch (err) {
      next(err);
    }
  });

  return function login(_x4, _x5, _x6) {
    return _ref2.apply(this, arguments);
  };
})();

let renew = (() => {
  var _ref3 = asyncToGenerator(function* (req, res, next) {
    try {
      const data = yield req.app.renew(req.jwt);
      res.ok(data);
    } catch (err) {
      next(err);
    }
  });

  return function renew(_x7, _x8, _x9) {
    return _ref3.apply(this, arguments);
  };
})();

let logout = (() => {
  var _ref4 = asyncToGenerator(function* (req, res, next) {
    try {
      const data = yield req.app.logout(req.jwt);
      res.ok(data);
    } catch (err) {
      next(err);
    }
  });

  return function logout(_x10, _x11, _x12) {
    return _ref4.apply(this, arguments);
  };
})();

const bearer_regex = /^Bearer\s*(.*)/;

// enhance the request with the jwt token if
// the Authorization header is present and holds a token
function extractJwtFromHeader(req, res, next) {
  const auth = (req.get('Authorization') || '').trim();

  if (auth) {
    const m = auth.match(bearer_regex);
    if (m) req.jwt = m[1];
  }

  next();
}

function HTTPError(status, message, code) {
  this.name = "HTTPError";
  this.status = status;
  this.message = message || http.STATUS_CODES[status];
  this.code = code || "EERROR";
  this.stack = new Error().stack;
}

HTTPError.prototype = Object.create(Error.prototype);

HTTPError.prototype.toString = function () {
  return `[${this.name}] ${this.status}: ${this.message}`;
};

function route(err, req, res, next) {
  if (!err) return next();
  if (!(err instanceof HTTPError)) {
    console.error(err.stack || err);
    err = new HTTPError(500);
  }

  res.status(err.status).json(_extends({
    error: true
  }, lodash.pick(err, "message", "status", "code")));
}

function defaultTransform(_ref) {
  let name = _ref.name,
      roles = _ref.roles,
      session = _ref.session,
      token = _ref.token,
      iat = _ref.iat,
      exp = _ref.exp;

  return {
    userCtx: { name, roles },
    session, token,
    issued: new Date(iat * 1000),
    expires: new Date(exp * 1000)
  };
}

var respUtils = function (_ref2) {
  var _ref2$transform = _ref2.transform;
  let transform = _ref2$transform === undefined ? defaultTransform : _ref2$transform;

  return function (req, res, next) {
    res.ok = function (data) {
      if (req.accepts("json")) {
        if (typeof transform === "function") {
          data = transform.call(req.app, data, this);
        }

        return res.json(_extends({
          ok: true
        }, lodash.omit(data, "ok")));
      }

      if (req.accepts("application/jwt")) {
        return res.type('application/jwt').send(data.token);
      }

      next(new HTTPError(406));
    };

    next();
  };
};

let defaultRefreshRoles = (() => {
  var _ref = asyncToGenerator(function* (_ref2, _ref3) {
    let name = _ref2.name,
        roles = _ref2.roles,
        token = _ref2.token;
    let baseUrl = _ref3.baseUrl;

    try {
      var _ref4 = yield superagent.get(`${baseUrl}/_users/org.couchdb.user:${name}`).set("Authorization", "Bearer " + token).accept("application/json");

      let body = _ref4.body;


      return body.roles;
    } catch (e) {
      return roles;
    }
  });

  return function defaultRefreshRoles(_x, _x2) {
    return _ref.apply(this, arguments);
  };
})();

var createAPI = function () {
  let couchOpts = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  let opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  const api = {};
  const baseUrl = couchOpts.baseUrl;
  const algorithms = opts.algorithms,
        expiresIn = opts.expiresIn,
        secret = opts.secret;
  var _opts$refreshRoles = opts.refreshRoles;
  const refreshRoles = _opts$refreshRoles === undefined ? defaultRefreshRoles : _opts$refreshRoles;

  const algorithm = Array.isArray(algorithms) ? algorithms[0] : "HS256";

  api.authenticate = (() => {
    var _ref5 = asyncToGenerator(function* (username, password) {
      try {
        if (typeof username !== "string" || !username) {
          throw new HTTPError(400, "Missing username.", "EBADINPUT");
        }

        if (typeof password !== "string" || !password) {
          throw new HTTPError(400, "Missing password.", "EBADINPUT");
        }

        var _ref6 = yield superagent.get(`${baseUrl}/_session`).accept("application/json").auth(username, password);

        let body = _ref6.body;


        return body;
      } catch (e) {
        let resp = e.response;
        if (!resp) throw e;

        if (resp.statusCode === 401) {
          throw new HTTPError(401, resp.body.reason, "EBADAUTH");
        } else {
          throw new HTTPError(resp.statusCode, resp.body.reason, "ECOUCH");
        }
      }
    });

    function authenticate(_x5, _x6) {
      return _ref5.apply(this, arguments);
    }

    return authenticate;
  })();

  api.generateToken = function generateToken(_ref7, session) {
    let name = _ref7.name,
        roles = _ref7.roles;

    const token = jwt.sign({ name, roles, session }, secret, { algorithm, expiresIn });
    const data = jwt.decode(token);
    data.token = token;
    return data;
  };

  api.generateSession = (() => {
    var _ref8 = asyncToGenerator(function* () {
      const sid = crypto.randomBytes(16).toString('hex');
      yield this.sessionStore.add(sid);
      return sid;
    });

    function generateSession() {
      return _ref8.apply(this, arguments);
    }

    return generateSession;
  })();

  api.validateToken = (() => {
    var _ref9 = asyncToGenerator(function* (token) {
      let opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      var _opts$ignoreExpiratio = opts.ignoreExpiration;
      let ignoreExpiration = _opts$ignoreExpiratio === undefined ? false : _opts$ignoreExpiratio;

      // decode data without verifying to check the session first

      const data = jwt.decode(token);
      if (!data) {
        throw new HTTPError(401, "Missing or invalid token.", "EBADTOKEN");
      }

      // ensure session id exists
      if (data.session) {
        const exists = yield this.sessionStore.exists(data.session);
        if (!exists) {
          throw new HTTPError(401, "Invalid session.", "EBADSESSION");
        }
      } else {
        throw new HTTPError(401, "Missing session id.", "EBADSESSION");
      }

      // verify the token
      try {
        jwt.verify(token, secret, {
          algorithms,
          ignoreExpiration
        });
      } catch (e) {
        if (e.name === "TokenExpiredError") {
          throw new HTTPError(401, "Expired token.", "EEXPTOKEN");
        } else if (e.name === "JsonWebTokenError") {
          throw new HTTPError(401, "Missing or invalid token.", "EBADTOKEN");
        }
      }

      // return the payload
      data.token = token;
      return data;
    });

    function validateToken(_x7) {
      return _ref9.apply(this, arguments);
    }

    return validateToken;
  })();

  api.refreshRoles = (() => {
    var _ref10 = asyncToGenerator(function* (data) {
      if (typeof refreshRoles === "function") {
        return yield refreshRoles.call(this, data, couchOpts);
      }

      return data.roles;
    });

    return function (_x9) {
      return _ref10.apply(this, arguments);
    };
  })();

  api.login = (() => {
    var _ref11 = asyncToGenerator(function* (username, password) {
      const response = yield this.authenticate(username, password);
      const session = yield this.generateSession();
      return this.generateToken(response.userCtx, session);
    });

    function login(_x10, _x11) {
      return _ref11.apply(this, arguments);
    }

    return login;
  })();

  api.logout = (() => {
    var _ref12 = asyncToGenerator(function* (token) {
      const data = yield this.validateToken(token, { ignoreExpiration: true });
      yield this.sessionStore.remove(data.session);
      return data;
    });

    function logout(_x12) {
      return _ref12.apply(this, arguments);
    }

    return logout;
  })();

  api.renew = (() => {
    var _ref13 = asyncToGenerator(function* (token) {
      const data = yield this.validateToken(token, { ignoreExpiration: true });
      data.roles = yield this.refreshRoles(data);
      return this.generateToken(data, data.session);
    });

    function renew(_x13) {
      return _ref13.apply(this, arguments);
    }

    return renew;
  })();

  return api;
};

const jsonParser = bodyParser.json();
const urlEncodedParser = bodyParser.urlencoded({ extended: true });

function createApp() {
  let opts = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var _opts$algorithms = opts.algorithms;
  let algorithms = _opts$algorithms === undefined ? ['HS256'] : _opts$algorithms;
  var _opts$session = opts.session;
  let session = _opts$session === undefined ? {} : _opts$session,
      couchdb = opts.couchdb;
  var _opts$endpoint = opts.endpoint;
  let endpoint = _opts$endpoint === undefined ? "/" : _opts$endpoint;
  var _opts$expiresIn = opts.expiresIn;
  let expiresIn = _opts$expiresIn === undefined ? '5m' : _opts$expiresIn,
      secret = opts.secret;
  var _opts$handleErrors = opts.handleErrors;
  let handleErrors = _opts$handleErrors === undefined ? true : _opts$handleErrors,
      transform = opts.transform,
      refreshRoles = opts.refreshRoles;


  if (!secret) throw new Error("Missing JWT secret.");

  // resolve session store
  const storeType = session.store;
  delete session.store;
  let createSessionStore;
  if (storeType === 'couch') {
    createSessionStore = require('./couch-store');
  } else if (storeType === 'empty') {
    createSessionStore = require('./empty-store');
  } else if (storeType === 'memory' || typeof storeType === 'undefined') {
    createSessionStore = require('./memory-store');
  } else if (typeof storeType === "string") {
    createSessionStore = require(storeType);
  } else if (typeof storeType === "function") {
    createSessionStore = storeType;
  } else {
    throw new Error("Expecting function or string for session storage.");
  }

  // create express app and parse options
  const couchOptions = getCouchDbOptions(couchdb);
  const app = express();
  app.disable('x-powered-by');

  // set a few helpers on the app
  const api = createAPI(couchOptions, {
    algorithms, expiresIn, secret, refreshRoles
  });

  lodash.forEach(api, (fn, name) => {
    if (typeof fn !== "function") return;
    app[name] = fn.bind(app);
  });

  // set up the session store
  let _setup = asyncToGenerator(function* () {
    let store = yield createSessionStore(session, couchOptions);
    app.sessionStore = store;
  })();
  app.setup = () => _setup;

  // app must be setup before routes can be used
  app.use((req, res, next) => {
    _setup.then(() => next(), next);
  });

  // mount the jwt auth logic
  app.route(endpoint).all(respUtils({ transform })).all(extractJwtFromHeader).delete(logout).get(info).post(jsonParser, urlEncodedParser, login).put(renew);

  // default error handling API
  // can be disabled when used with a greater express app
  if (handleErrors) {
    // default everything else to a 404
    app.use((req, res, next) => next(new HTTPError(404)));

    // mount the error route
    app.use(route);
  }

  return app;
}

createApp.HTTPError = HTTPError;

module.exports = createApp;
